"""
Schema for some basic data types
--------------------------------
"""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class Vector3D(BaseModel):
    """
    Vector in three dimensions in space.
    """

    x: float = 0
    "Value in x direction."
    y: float = 0
    "Value in y direction."
    z: float = 0
    "Value in z direction."
    model_config = ConfigDict(extra="forbid")
    "Configuration for :py:class:`pydantic.BaseModel`."


class Slice(BaseModel):
    """
    Slice of an array.

    :see: :py:class:`slice`
    """

    axis: int = -1
    "Axis of the slice."
    start: int | None = None
    "Start of the slice."
    end: int | None = None
    "End of the slice."
    step: int | None = None
    "Step of the slice."
    model_config = ConfigDict(extra="forbid")
    "Configuration for :py:class:`pydantic.BaseModel`."

    def __call__(self) -> slice:
        """
        Return a slice object.

        :return: Slice object.
        :see: :py:class:`slice`
        """
        return slice(self.start, self.end, self.step)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest import mock

import matplotlib.pyplot as plt
import numpy as np
import yaml

from pigreads.plot import LiveView, movie, movie_progress_updater, movie_wrapper
from pigreads.schema import Simulation

yaml_data = """
pigreads: 1
Nfr: 3
Nt: 2
Nz: 1
Ny: 50
Nx: 100
dt: 0.1
dz: 1.0
dy: 0.02
dx: 0.02
diffusivity: 0.1
models: aliev1996simple
init:
    u: 0.3
"""
sim = Simulation(**yaml.safe_load(yaml_data))
states = sim.run()


@dataclass
class MockClickEvent:
    inaxes: Any = None
    xdata: Any = None
    ydata: Any = None


def test_movie():
    with TemporaryDirectory() as tempdir:
        path = str(Path(tempdir) / "movie.mp4")
        movie(path, states[:, 0, :, :, 0])
        assert Path(path).exists()


def test_movie_wrapper():
    with TemporaryDirectory() as tempdir:
        path = str(Path(tempdir) / "movie.mp4")
        movie_wrapper({"path": path, "frames": states[:, 0, :, :, 0]})


def test_movie_progress_dict():
    with TemporaryDirectory() as tempdir:
        path = str(Path(tempdir) / "movie.mp4")
        progress = {path: 0}
        movie(path, states[:, 0, :, :, 0], progress_dict=progress)
        assert progress[path] == sim.Nfr


def test_movie_parallel():
    # Can not test the actual parallelism in this test.
    # Just check if the function runs without error.

    with (
        TemporaryDirectory() as tempdir,
        mock.patch("multiprocessing.Manager") as MockManager,
        mock.patch("multiprocessing.Pool") as MockPool,
        mock.patch("multiprocessing.Process") as MockProcess,
        mock.patch("subprocess.Popen") as MockPopen,
    ):
        mock_manager_instance = mock.MagicMock()
        mock_manager_instance.dict.side_effect = lambda d: d
        MockManager.return_value.__enter__.return_value = mock_manager_instance

        mock_pool_instance = mock.MagicMock()
        mock_async_result = mock.MagicMock()
        mock_async_result.get.return_value = None
        mock_pool_instance.starmap_async.return_value = mock_async_result
        MockPool.return_value.__enter__.return_value = mock_pool_instance

        mock_process_instance = mock.MagicMock()
        MockProcess.return_value = mock_process_instance

        mock_popen_instance = mock.MagicMock()
        mock_popen_instance.communicate.return_value = (b"", b"")
        mock_popen_instance.returncode = 0
        MockPopen.return_value = mock_popen_instance

        path = str(Path(tempdir) / "movie.mp4")
        movie(path, states[:, 0, :, :, 0], parallel=2)

        mock_process_instance.start.assert_called_once()
        mock_process_instance.join.assert_called_once()
        mock_pool_instance.starmap_async.assert_called_once()
        MockPopen.assert_called_once()


def test_movie_progress_updater():
    # pylint: disable=all

    class TestProgressDict(dict[str, int]):
        def __init__(self) -> None:
            super().__init__()
            self.counter: int = 0

        def values(self) -> Any:
            self.counter += 1
            return [self.counter // 3]

    progress_dict: TestProgressDict = TestProgressDict()

    with mock.patch("time.sleep", return_value=None):
        movie_progress_updater("none", 3, progress_dict)

    assert progress_dict.counter == 6


def test_liveview_show():
    fig, ax = plt.subplots()
    with (
        mock.patch.object(fig, "show"),
        mock.patch("matplotlib.pyplot.subplots") as mock_subplots,
    ):
        mock_subplots.return_value = (fig, ax)
        liveview = LiveView(sim)
        liveview.update(states, 0)
        liveview.close()


def test_liveview_clicks():
    with TemporaryDirectory() as tempdir:
        liveview = LiveView(
            sim=sim,
            path=str(Path(tempdir) / "live.png"),
            click_radius=0.2,
            click_value=123,
        )

        liveview.on_press(MockClickEvent(inaxes=liveview.ax, xdata=2, ydata=3))
        assert liveview.click_location == (3, 2)
        assert liveview.click_location_prev is None

        liveview.on_motion(MockClickEvent(inaxes=liveview.ax, xdata=4, ydata=1))
        assert liveview.click_location == (1, 4)
        assert liveview.click_location_prev is None

        liveview.on_release(MockClickEvent(inaxes=liveview.ax, xdata=None, ydata=None))
        assert liveview.click_location is None
        assert liveview.click_location_prev is None  # type: ignore[unreachable]

        liveview.close()


def test_liveview_clicks_none():
    with TemporaryDirectory() as tempdir:
        liveview = LiveView(
            sim=sim,
            path=str(Path(tempdir) / "live.png"),
        )

        liveview.on_press(MockClickEvent(inaxes=liveview.ax, xdata=2, ydata=3))
        assert liveview.click_location is None
        assert liveview.click_location_prev is None

        liveview.on_motion(MockClickEvent(inaxes=liveview.ax, xdata=4, ydata=1))
        assert liveview.click_location is None
        assert liveview.click_location_prev is None

        liveview.on_release(MockClickEvent(inaxes=liveview.ax, xdata=None, ydata=None))
        assert liveview.click_location is None
        assert liveview.click_location_prev is None

        liveview.close()


def test_liveview_draw():
    with TemporaryDirectory() as tempdir:
        liveview = LiveView(
            sim=sim,
            path=str(Path(tempdir) / "live.png"),
            click_radius=0.2,
            click_value=123,
        )
        liveview.click_location = (0.3, 0.8)
        liveview.click_location_prev = (0.7, 1.2)
        frame = np.copy(states[-1, 0, :, :, 0])
        liveview.draw_thick_line(frame)
        assert np.sum(frame) == 111793.836
        assert liveview.click_location_prev == (0.3, 0.8)



"""
Plots and movies
----------------
"""

from __future__ import annotations

import multiprocessing
import subprocess
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FFMpegWriter
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.image import AxesImage

from pigreads import Models
from pigreads.progress import PROGRESS_ITERS
from pigreads.schema import Simulation


def get_iz_iv(
    sim: Simulation,
    models: Models,
    variable: str | int | None,
    index_z: int | None,
) -> tuple[int, int]:
    """
    Interpret z index and variable index from command line arguments.

    :param sim: Simulation object.
    :param models: Models object.
    :param variable: Variable name or index.
    :param index_z: Index in z direction.
    :return: Tuple of z index and variable index.
    """

    iz: int = sim.Nz // 2 if index_z is None else index_z
    iv: int

    try:
        iv = int(variable or 0)
        assert iv >= 0, "Invalid variable index"
        assert iv < models.Nv, "Invalid variable index"

    except ValueError:
        iv = 0 if variable is None else sim.varidx(models)[str(variable)]

    assert iz >= 0, "Invalid z index"
    assert iz < sim.Nz, "Invalid z index"

    return iz, iv


def imshow_defaults(
    array: np.ndarray[Any, Any] | None = None,
    sim: Simulation | None = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """
    Set default imshow arguments.

    :param array: Array to display.
    :param sim: Simulation object.
    :param kwargs: Additional arguments.
    :return: Dictionary of arguments for
             :py:func:`matplotlib.pyplot.imshow`.
    """

    if "origin" not in kwargs:
        kwargs["origin"] = "lower"

    if "interpolation" not in kwargs:
        kwargs["interpolation"] = "none"

    if array is not None:
        if "vmin" not in kwargs:
            kwargs["vmin"] = np.nanmin(array)

        if "vmax" not in kwargs:
            kwargs["vmax"] = np.nanmax(array)

    if sim is not None and "extent" not in kwargs:
        kwargs["extent"] = (
            -0.5 * sim.dx,
            (sim.Nx + 0.5) * sim.dx,
            -0.5 * sim.dy,
            (sim.Ny + 0.5) * sim.dy,
        )

    return kwargs


def plot_frame(
    ax: Axes,
    frame: np.ndarray[Any, Any],
    xlabel: str = "x",
    ylabel: str = "y",
    vlabel: str = "",
    title: str = "",
    **kwargs: Any,
) -> tuple[AxesImage, Colorbar]:
    """
    Display a frame as an image.

    :param ax: Axes object.
    :param xlabel: Label for the x-axis.
    :param ylabel: Label for the y-axis.
    :param vlabel: Colorbar label.
    :param title: Title of the plot.
    :param frame: Frame to display.
    :param kwargs: Passed to :py:func:`matplotlib.pyplot.imshow`.

    :return: Image and colorbar objects.
    """
    assert frame.ndim == 2
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    im = ax.imshow(frame, **imshow_defaults(array=frame, **kwargs))
    cbar = plt.colorbar(im)
    cbar.set_label(vlabel)
    return im, cbar


def movie(
    path: str,
    frames: np.ndarray[Any, Any],
    dpi: int = 180,
    fps: int = 15,
    tlables: list[str] | None = None,
    progress: str = "none",
    progress_dict: dict[str, int] | None = None,
    parallel: int = 1,
    **kwargs: Any,
) -> None:
    """
    Render a chunk of frames to a movie file, with optional parallelization.

    :param path: Path to write the movie file to.
    :param frames: Array of frames.
    :param dpi: Dots per inch.
    :param fps: Frames per second.
    :param tlables: List of time labels.
    :param progress: Progress bar type.
    :param progress_dict: Dictionary to store progress.
    :param parallel: Number of processes (default 1, 0 to use all CPUs).
    :param kwargs: Passed to :py:func:`plot_frame`.
    """
    assert frames.ndim == 3

    if parallel == 1:
        fig, ax = plt.subplots(dpi=dpi)
        writer = FFMpegWriter(fps=fps)
        tlables = [f"{i}" for i, _ in enumerate(frames)] if tlables is None else tlables
        im, _ = plot_frame(ax, frames[0], **imshow_defaults(array=frames, **kwargs))

        prog = PROGRESS_ITERS[progress]
        with writer.saving(fig, path, fig.dpi):
            for (i, frame), tlabel in zip(enumerate(prog(frames)), tlables):
                ax.set_title(tlabel)
                im.set_data(frame)
                writer.grab_frame()
                if progress_dict is not None:
                    progress_dict[path] = i + 1
        return

    Np = multiprocessing.cpu_count() if parallel == 0 else parallel  # pylint: disable=invalid-name
    Nfr = len(frames)  # pylint: disable=invalid-name
    Nfrp = (Nfr + Np - 1) // Np  # pylint: disable=invalid-name
    chunks = [slice(i, min(i + Nfrp, Nfr)) for i in range(0, Nfr, Nfrp)]

    with TemporaryDirectory() as temp:
        paths = [str(Path(temp) / f"{i}.mp4") for i, _ in enumerate(chunks)]

        with multiprocessing.Manager() as manager:
            assert progress_dict is None
            progress_dict_ = manager.dict(dict.fromkeys(paths, 0))
            tasks = [
                [
                    {
                        "path": path,
                        "frames": frames[chunk],
                        "tlables": tlables[chunk] if tlables else None,
                        "progress_dict": progress_dict_,
                        "parallel": 1,
                        "dpi": dpi,
                        "fps": fps,
                        **kwargs,
                    }
                ]
                for path, chunk in zip(paths, chunks)
                if chunk.start < chunk.stop
            ]

            progress_proc = multiprocessing.Process(
                target=movie_progress_updater,
                args=(progress, Nfr, progress_dict_),
            )
            progress_proc.start()

            with multiprocessing.Pool(processes=Np) as pool:
                pool.starmap_async(movie_wrapper, tasks).get()

            progress_proc.join()

        pathlist = Path(temp) / "files.txt"
        with pathlist.open("w") as f:
            for p in paths:
                f.write(f"file '{p}'\n")

        proc = subprocess.Popen(  # pylint: disable=consider-using-with
            [
                "ffmpeg",
                "-y",
                "-f",
                "concat",
                "-safe",
                "0",
                "-i",
                pathlist,
                "-c",
                "copy",
                str(path),
            ],
            stderr=subprocess.PIPE,
        )

        _, stderr = proc.communicate()
        assert proc.returncode == 0, stderr.decode()


def movie_wrapper(kwargs: dict[str, Any]) -> None:
    """
    Wrapper for movie to allow for multiprocessing.

    :param kwargs: Keyword arguments for :py:func:`movie`.
    """
    return movie(**kwargs)


def movie_progress_updater(
    progress: str, total: int, progress_dict: dict[str, int]
) -> None:
    """
    Update the progress bar for a movie.

    :param progress: Progress bar type.
    :param total: Total number of frames.
    :param progress_dict: Dictionary to store progress.
    """
    prog = PROGRESS_ITERS[progress]
    for i in prog(range(total)):
        while sum(progress_dict.values()) < i:
            time.sleep(0.1)


class LiveView:
    """
    Live view for a simulation.

    Plot a single variable at a fixed z index in an interactive window or save
    it to a file.

    :param sim: Simulation object.
    :param variable: Variable name or index.
    :param index_z: Index in z direction.
    :param dpi: Dots per inch.
    :param path: File path.
    :param click_radius: Radius of click region.
    :param click_value: Value to add continuously while clicking.
    :param style: Style sheet,
                  passed to :py:func:`matplotlib.style.use`,
                  for example ``dark_background``.
    :param kwargs: Additional arguments to pass to :py:func:`plot_frame` and
                   :py:func:`matplotlib.pyplot.imshow`.

    :ivar sim: Simulation object.
    :ivar fig: Figure object.
    :ivar ax: Axes object.
    :ivar models: Models object.
    :ivar iz: Index in z direction.
    :ivar iv: Index of variable.
    :ivar click_radius: Radius of click region.
    :ivar click_value: Value to add continuously while clicking.
    :ivar click_location: Location of the current click.
    :ivar click_location_prev: Location of the previous click.
    :ivar mouse_pressed: Flag indicating whether the mouse is pressed.
    :ivar kwargs: Additional arguments.
    :ivar path: File path.
    :ivar im: Image object.
    :ivar cbar: Colorbar object.
    """

    def __init__(
        self,
        sim: Simulation,
        variable: str | int | None = None,
        index_z: int | None = None,
        dpi: int | None = None,
        path: Path | str | None = None,
        click_radius: float | None = None,
        click_value: float | None = None,
        style: str | None = None,
        **kwargs: Any,
    ) -> None:
        if style is not None:
            plt.style.use(style)

        self.sim = sim
        self.kwargs = kwargs
        self.fig, self.ax = plt.subplots(dpi=dpi)
        plt.tight_layout()
        self.models = sim.prepare_models()
        self.iz, self.iv = get_iz_iv(sim, self.models, variable, index_z)
        self.kwargs = {
            "vlabel": list(self.models.available[self.models[0].key].variables)[
                self.iv
            ],
            **imshow_defaults(sim=sim, **kwargs),
        }
        self.path = path
        self.im: AxesImage | None = None
        self.cbar: Colorbar | None = None

        if self.path is None:
            self.fig.show()

        self.click_radius = click_radius
        self.click_value = click_value
        self.click_location: tuple[float, float] | None = None
        self.click_location_prev: tuple[float, float] | None = None
        self.mouse_pressed = False

        self.fig.canvas.mpl_connect("button_press_event", self.on_press)
        self.fig.canvas.mpl_connect("motion_notify_event", self.on_motion)
        self.fig.canvas.mpl_connect("button_release_event", self.on_release)

    def on_press(self, event: Any) -> None:
        """
        Handle mouse button press: store click location and set press flag.
        """
        if (
            self.click_radius is None
            or self.click_value is None
            or event.inaxes != self.ax
            or event.xdata is None
            or event.ydata is None
        ):
            return
        self.click_location = (event.ydata, event.xdata)
        self.mouse_pressed = True

    def on_motion(self, event: Any) -> None:
        """
        Handle mouse motion: draw a thick line between previous and current locations.
        """
        if (
            self.click_radius is None
            or self.click_value is None
            or not self.mouse_pressed
            or event.inaxes != self.ax
            or event.xdata is None
            or event.ydata is None
        ):
            return
        self.click_location = (event.ydata, event.xdata)

    def on_release(self, _: Any) -> None:
        """
        Handle mouse button release: clear press flag.
        """
        if self.click_radius is None or self.click_value is None:
            return
        self.mouse_pressed = False
        self.click_location = None
        self.click_location_prev = None

    def draw_thick_line(self, frame: np.ndarray[Any, Any]) -> None:
        """
        Draw a thick line between the previous and new locations with the given radius.
        """
        if (
            self.click_radius is None
            or self.click_value is None
            or self.click_location is None
        ):
            return

        y1, x1 = self.click_location
        y2, x2 = self.click_location

        if self.click_location_prev is not None:
            y2, x2 = self.click_location_prev

        dx, dy = x2 - x1, y2 - y1

        y, x = np.meshgrid(
            self.sim.dy * np.arange(self.sim.Ny),
            self.sim.dx * np.arange(self.sim.Nx),
            indexing="ij",
        )

        distance = np.abs(+dy * x - dx * y + x2 * y1 - y2 * x1) / np.linalg.norm(
            (dy, dx), axis=0
        )

        t = ((x - x1) * dx + (y - y1) * dy) / (dx**2 + dy**2)

        mask = (distance <= self.click_radius) & (t >= 0) & (t <= 1)

        for y_, x_ in [(y1, x1), (y2, x2)]:
            mask[np.linalg.norm([y - y_, x - x_], axis=0) <= self.click_radius] = True

        frame[mask] += self.click_value
        self.click_location_prev = self.click_location

    def update(self, states: np.ndarray[Any, Any], ifr: int) -> None:
        """
        Update the live view.

        :param states: Array of states with shape (Nfr, Nz, Ny, Nx, Nv).
        :param ifr: Index of frame.
        """

        frame = states[ifr, self.iz, :, :, self.iv]

        self.draw_thick_line(frame)

        if self.im is None:
            self.im, self.cbar = plot_frame(
                self.ax, frame, **imshow_defaults(array=frame, **self.kwargs)
            )

        else:
            assert self.cbar is not None
            vmin, vmax = self.im.get_clim()
            vmin = self.kwargs.get("vmin", np.nanmin([vmin, np.nanmin(frame)]))
            vmax = self.kwargs.get("vmax", np.nanmax([vmax, np.nanmax(frame)]))
            self.im.set_data(frame)
            self.im.set_clim(vmin, vmax)
            self.cbar.mappable.set_clim(vmin, vmax)
            self.cbar.update_normal(self.im)

        self.ax.set_title(
            f"frame {ifr}/{self.sim.Nfr}, t = {ifr * self.sim.Nt * self.sim.dt:.2f}"
        )
        if self.path is None:
            self.fig.canvas.draw_idle()
            self.fig.canvas.flush_events()
        else:
            self.fig.savefig(str(self.path))

    def close(self) -> None:
        """
        Close the opened figure.
        """
        plt.close(self.fig)



from __future__ import annotations

from typing import Any

import numpy as np
import pytest

import pigreads as pig
from pigreads.schema.model import ModelDefinition


def test_str():
    pig.Models("marcotte2017dynamical")


def test_add():
    key = "marcotte2017dynamical"
    models = pig.Models(
        [
            (key, {}),
            (key, {"eps": 1}),
        ]
    )
    models.add(key)
    models.add(key, eps=2)
    assert models[0] == models[2]
    assert models[1].parameters["eps"] == 1
    assert models[3].parameters["eps"] == 2


def test_len():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    models.add(key)
    models.add(key)
    assert len(models) == 3


def test_Nv():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    assert models.Nv == 2


def test_Np():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    assert len(models[0].parameters) == 5


def test_set_param():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    models[0].parameters["eps"] *= 2


def test_del_param():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    with pytest.raises(NotImplementedError):
        del models[0].parameters["eps"]


def test_repr_params():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    repr(models[0].parameters)


def test_str_params():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    str(models[0].parameters)


def test_resting_states():
    models = pig.Models()
    models.add("gray1983autocatalytic")
    models.add("courtemanche1998ionic")
    inhom = np.array([[[0, 1], [2, 3]]], dtype=int)
    states = models.resting_states(inhom, Nframes=2)
    assert states.shape == (2, *inhom.shape, models.Nv)
    assert np.all(np.isnan(states[0, 0, 0, 0, :]))
    assert np.all(np.isnan(states[:, inhom == 0, :]))
    assert states[0, 0, 0, 1, 0] == 1
    assert states[0, 0, 0, 1, 1] == 0
    assert abs(-81.18 - states[0, 0, 1, 0, 0]) < 1e-6
    assert abs(11.17 - states[0, 0, 1, 0, 1]) < 1e-6
    assert states[0, 0, 1, 1, 0] == 1
    assert states[0, 0, 1, 1, 1] == 0
    assert np.all(np.isnan(states[1, inhom != 0, :]))


def test_sim0d():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    for modelname in pig.Models().available:
        models = pig.Models(modelname)
        states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
            np.float32
        )
        weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
        states[1] = models.run(inhom, weights, states[0], Nt=2, dt=1e-10)


def test_define():
    pig.Models.available["fitzhugh1961impulses"] = ModelDefinition(
        name="FitzHugh 1961 & Nagumo 1962",
        description="A 2D simplification of the Hodgkin-Huxley model.",
        dois=[
            "https://doi.org/10.1016/S0006-3495(61)86902-6",
            "https://doi.org/10.1109/JRPROC.1962.288235",
        ],
        variables={"u": 1.2, "v": -0.625},
        diffusivity={"u": 1.0},
        parameters={"a": 0.7, "b": 0.8, "c": 3.0, "z": 0.0},
        code="""
            *_new_u = u + dt * (v + u - u*u*u/3 + z + _diffuse_u);
            *_new_v = v + dt * (-(u - a + b*v)/c);
        """,
    )

    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    models = pig.Models("fitzhugh1961impulses")

    states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    states[1] = models.run(inhom, weights, states[0], Nt=100, dt=0.004)
    assert np.allclose(states[1], [1.1996392, -0.6249765])


def test_block_size():
    models = pig.Models()
    assert models.block_size == [1, 8, 8]
    models.block_size = [8, 8, 8]
    assert models.block_size == [8, 8, 8]
